package org.openxma.dsl.pom.naming;

import java.util.Collections;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.xtext.naming.DefaultDeclarativeQualifiedNameProvider;
import org.eclipse.xtext.naming.QualifiedName;
import org.eclipse.xtext.util.IResourceScopeCache;
import org.eclipse.xtext.util.PolymorphicDispatcher;
import org.eclipse.xtext.util.Strings;
import org.eclipse.xtext.util.Tuples;
import org.openxma.dsl.core.model.Model;
import org.openxma.dsl.dom.model.Attribute;
import org.openxma.dsl.dom.model.FeatureReference;
import org.openxma.dsl.dom.model.PresentableFeature;
import org.openxma.dsl.pom.PomPackage;
import org.openxma.dsl.pom.PomSingletons;
import org.openxma.dsl.pom.formatting.PomModelSerializer;
import org.openxma.dsl.pom.model.Component;
import org.openxma.dsl.pom.model.CustomizeableField;
import org.openxma.dsl.pom.model.DataObjectVariable;
import org.openxma.dsl.pom.model.FieldFeature;
import org.openxma.dsl.pom.model.FieldReference;
import org.openxma.dsl.pom.model.IComposite;
import org.openxma.dsl.pom.model.IField;
import org.openxma.dsl.pom.model.Image;
import org.openxma.dsl.pom.model.Label;
import org.openxma.dsl.pom.model.ReferencedXMAComposite;
import org.openxma.dsl.pom.model.ReferencedXMAPage;
import org.openxma.dsl.pom.model.TabFolder;
import org.openxma.dsl.pom.model.TabPage;
import org.openxma.dsl.pom.model.Table;
import org.openxma.dsl.pom.model.TableColumn;
import org.openxma.dsl.pom.model.TableCombo;
import org.openxma.dsl.pom.model.TabulatorPosition;

import at.spardat.xma.guidesign.GuidesignPackage;
import at.spardat.xma.guidesign.XMAComponent;

import com.google.common.base.Function;
import com.google.inject.Inject;
import com.google.inject.Provider;


public class PomNameProvider extends DefaultDeclarativeQualifiedNameProvider {
	static public PomNameProvider INSTANCE = new PomNameProvider();
	static GuiDesignNameFunction guiDesignNameFunction = new GuiDesignNameFunction();

	//@Inject
	//PomNameFunction pomNameFunction;

	public static PomNameFunction getPomNameFunction() {
		return PomSingletons.getPomNameFunction();
	}	
	
	public static Function<EObject, String> XMA_ID = new Function<EObject, String>() {
		public String apply(EObject from) {
			return PomNameProvider.getXmaName(from);
		}
	};

	public static String getXmaName(EObject from) {
		if (from instanceof Component) {
			XMAComponent xmaComponent = ((Component) from).getXmaComponent();
			return xmaComponent == null ? null : getXmaName(xmaComponent);
		} else if (from instanceof ReferencedXMAComposite) {
			return ((ReferencedXMAComposite) from).getXmaComposite() == null ? null
					: getXmaName(((ReferencedXMAComposite) from).getXmaComposite());
		} else if (from instanceof ReferencedXMAPage) {
			return ((ReferencedXMAPage) from).getXmaPage() == null ? null : getXmaName(((ReferencedXMAPage) from)
					.getXmaPage());
		} else if (from instanceof TabulatorPosition) {
			return ((TabulatorPosition) from).getName();
		} else if (from instanceof FieldReference) {
			FieldReference ref = (FieldReference) from;
			return "fieldReference_"
					+ (ref.getObject().getName() + PomModelSerializer.getFeatureDefinitionString(ref.getFeature()));
		} else if (from instanceof CustomizeableField) {
			CustomizeableField f = (CustomizeableField) from;
			return composeDefaultWidgetNameForField(f);
		} else if (from instanceof TableColumn) {
			TableColumn f = (TableColumn) from;
			return composeDefaultWidgetNameForColumn(f);
		} else if (from instanceof IField) {
			throw new RuntimeException("Method should not be reached");
		} else if (from == null) {
			return null;
		} 
		/*
		 * else if (from instanceof Content) { return
		 * getXmaName(from.eContainer()) + "-content"; }
		 */
		QualifiedName qualifiedName = INSTANCE.getFullyQualifiedName(from);
		if (qualifiedName == null) {
			return null;
		}
		return qualifiedName.getLastSegment();
	}

	/**
	 * Compose a default name for a GuiElement. This is intended to be used to
	 * determine a default name in case the name property is not defined (and
	 * therefore null).
	 * 
	 * @param obj
	 * @return
	 */
	public static String composeDefaultNameForGuiElement(EObject obj) {
		String typeName = "";
		if (obj instanceof IComposite) {
			typeName = "composite";
		} else if (obj instanceof Label) {
			typeName = "label";
		} else if (obj instanceof Image) {
			typeName = "image";
		} else if (obj instanceof TabFolder) {
			typeName = "tabfolder";
		} else if (obj instanceof TabPage) {
			typeName = "notebookpage";
		}
		String postFix = "";
		while (obj != null) {
			EObject parent = obj.eContainer();
			if (parent instanceof Model) {
				return null;
			}
			int index = parent.eContents().indexOf(obj);
			postFix += "_" + String.valueOf(index);
			String parentName = getXmaName(parent);
			if (parentName != null) {
				String defaultName = parentName + "_" + typeName + postFix;
				return defaultName;
			}
			obj = parent;
		}
		return null;
	}
				
	/**
	 * Compose the default widget name for a field
	 */
	public static String composeDefaultWidgetNameForColumn(TableColumn column) {
	    String name = null;
	    if (column.eContainer() instanceof Table)
	        name = ((Table)column.eContainer()).getName();
	    if (column.eContainer() instanceof TableCombo)
	        name = ((TableCombo)column.eContainer()).getName();
	    
		if (column.getFeature() != null) {
			if (column.getFeature().getAttributeHolder() == null) {
				throw new IllegalArgumentException("The attribute of a feature can not be found");
			}
			return name+"_"+column.getFeature().getAttributeHolder().getCalculatedName();
		} else {
			return name+"_"+column.getColumnName();
		}			
	}	

	/**
	 * Compose the default widget name for a field
	 */
	public static String composeDefaultWidgetNameForField(IField f) {
		if (f.getObject() == null)
			return null;
		return composeDefaultWidgetNameForAttribute(f.getObject(), f.getFeature());
	}

	private static String composeDefaultWidgetNameForAttribute(DataObjectVariable dov, FieldFeature fieldFeature) {
		StringBuilder buf = new StringBuilder();
		buf.append(dov.getName());
		buf.append(composeDefaultWidgetNameForFeature(fieldFeature));
		return buf.toString();
	}

	public static String composeDefaultWidgetNameForAttribute(DataObjectVariable dov, PresentableFeature feature) {
		return dov.getName() + "_" + feature.getCalculatedName();
	}

	private static String composeDefaultWidgetNameForFeature(FieldFeature fieldFeature) {
		PresentableFeature feature = fieldFeature.getFeature();
		String name = null;
		if (feature instanceof Attribute) {
			name = feature.getName();
		} 
		if (name == null) {
			name = getPomNameFunction().apply(fieldFeature);
		}
		if (name == null && feature.getAttribute() != null) {
			name = feature.getAttribute().getName();
		}
		StringBuilder buf = new StringBuilder("_" + name);
		if (fieldFeature.getSubFeature() != null) {
			buf.append(composeDefaultWidgetNameForFeature(fieldFeature.getSubFeature()));
		}
		return buf.toString();
	}

	public static String getScrolledName(String name) {
		return "scrolled" + name;
	}
	
	/**
	 * Get the name for the containing Composite of a Table which has a
	 * paging control attached.
	 * 
	 * @param name
	 * @return
	 */
	public static String getTableCompositeName(String name) {
		return name+"_composite";
	}	

	@Inject
	private IResourceScopeCache cache = IResourceScopeCache.NullImpl.INSTANCE;

    private PolymorphicDispatcher<QualifiedName> qualifiedName = new PolymorphicDispatcher<QualifiedName>(
            "qualifiedName", 1, 1, Collections.singletonList(this),
            PolymorphicDispatcher.NullErrorHandler.<QualifiedName> get()) {
        @Override
        protected QualifiedName handleNoSuchMethod(Object... params) {
            return null;
        }
    };

	
	public QualifiedName getFullyQualifiedName(final EObject obj) {
        return cache.get(Tuples.pair(obj, "fqn"), obj.eResource(), new Provider<QualifiedName>(){

            public QualifiedName get() {
                EObject temp = obj;
                QualifiedName qualifiedNameFromDispatcher = qualifiedName.invoke(temp);
                if (qualifiedNameFromDispatcher!=null)
                    return qualifiedNameFromDispatcher;
                String name = getResolver(obj).apply(temp);
                if (Strings.isEmpty(name))
                    return null;
                QualifiedName qualifiedNameFromConverter = getConverter().toQualifiedName(name);
                while (temp.eContainer() != null) {
                    temp = temp.eContainer();
                    QualifiedName parentsQualifiedName = getFullyQualifiedName(temp);
                    if (parentsQualifiedName != null)
                        return parentsQualifiedName.append(qualifiedNameFromConverter);
                }
                return qualifiedNameFromConverter;
            }
        });
    }
	
	public QualifiedName qualifiedName(FeatureReference featureReference) {
		String name = featureReference.getName();
		if (name==null) {
			name = featureReference.getAttribute().getName();
		}
		return QualifiedName.create(name);
	}
	
	public QualifiedName qualifiedName(PresentableFeature presentableFeature) {
		return QualifiedName.create(presentableFeature.getName());
	}
	
	protected Function<EObject, String> getResolver(EObject eObject) {
		if (null != GuidesignPackage.eINSTANCE.getEClassifier(eObject.eClass().getName())) {
			return guiDesignNameFunction;
		} else if (null != PomPackage.eINSTANCE.getEClassifier(eObject.eClass().getName())) {
			return getPomNameFunction();
		}
		return super.getResolver();
	}

}
